// SPDX-License-Identifier: MIT OR LGPL-2.0-or-later
// SPDX-FileCopyrightText: 2025 Philip Chimento <philip.chimento@gmail.com>
const {get2ndNumber} = await import('./sourcemap-number-module.js');
debugger;
get2ndNumber([{ n: 1 }, { n: null }]);
