var a = { n: null };
var b = a.n.toString(42);
//# sourceMappingURL=data:application/json;base64,eyJ2ZXJzaW9uIjozLCJmaWxlIjoibm9Nb2R1bGUuanMiLCJzb3VyY2VSb290IjoiIiwic291cmNlcyI6WyJub01vZHVsZS50cyJdLCJuYW1lcyI6W10sIm1hcHBpbmdzIjoiQUFJQSxJQUFJLENBQUMsR0FBZ0IsRUFBQyxDQUFDLEVBQUUsSUFBSSxFQUFDLENBQUM7QUFDL0IsSUFBTSxDQUFDLEdBQUcsQ0FBQyxDQUFDLENBQUMsQ0FBQyxRQUFRLENBQUMsRUFBRSxDQUFDLENBQUMiLCJzb3VyY2VzQ29udGVudCI6WyJpbnRlcmZhY2UgRmFuY3lOdW1iZXIge1xuICAgIG46IG51bWJlcjtcbn1cblxubGV0IGE6IEZhbmN5TnVtYmVyID0ge246IG51bGx9O1xuY29uc3QgYiA9IGEubi50b1N0cmluZyg0Mik7XG5cbiJdfQ==